from netsc import Server
from .other import serialize_card, deserialize_card

class GameServer(Server):
    bind_addr = ('', 8660)
    attrs = ['hand', 'name']
    def adj_args_start(self, game):
        self.wrapped = game
        return None,
    def adj_args_play(self, *args, **kwargs):
        args = list(args)
        args[0] = serialize_card(args[0])
        args[1] = None
        return tuple(args)
    def deadj_return_play(self, card):
        if card is None:
            return card
        return deserialize_card(card)
    def adj_args_can_play_card(self, card1, card2):
        return serialize_card(card1), serialize_card(card2)
    def adj_args_can_play(self, card):
        return serialize_card(card),
    def adj_args_remove_from_hand(self, card):
        return serialize_card(card),
    def deadj_return___getattribute__(self, value):
        if isinstance(value, tuple):
            return [deserialize_card(x) for x in value]
        else: return value


from ._color import *
from ._mods import *
from ._core import *
from ._card import *
from .lang import *

class Player:
    __slots__ = ['hand', 'name']
    def __init__(self, card_count=7):
        self.hand = []
        self.draw(card_count)
        self.name = 'Player'
    def start(self, game): pass
    def draw(self, count):
        self.hand.extend(draw(count))
    def remove_from_hand(self, card):
        self.hand.remove(card)
    def score(self):
        return tally(self.hand)
    def play(self, current_card, game):
        def do():
            while True:
                card = self._play(current_card)
                if self.can_play_card(current_card, card):
                    game.display_message("%s has %i cards" % (self.name, len(self.hand)-1))
                    return card
        if self.can_play(current_card):
            return do()
        else:
            game.display_message("%s couldn't play and had to draw a card" % self.name)
            self.draw(1)
            if self.can_play(current_card):
                return do()
            else:
                game.display_message("%s still couldn't play and had to be skipped" % self.name)
                return None
    def _play(self, current_card):
        for card in self.hand:
            if self.can_play_card(current_card, card):
                return card
        # return self.hand[random.randint(0, len(self.hand)-1)]
    def ask(self, q, t=str, limits=()):
        while True:
            value = self._ask(q, t)
            value, valid = self._validate_ask(value, t, limits)
            if valid: return value
    def _ask(self, q, t):
        if issubclass(t, Color):
            return random.choice(list(COLOR_SET))
        elif issubclass(t, str):
            return random.choice(string.printable)
        elif issubclass(t, bool):
            return bool(random.randint(0, 1))
    def _validate_ask(self, v, t, limits):
        ret = ()
        if isinstance(v, t): ret = v, True
        elif issubclass(t, Color):
            for color in COLOR_SET:
                if color.code == str(v[0]):
                    ret = color, True
                    break
            else:
                ret = v, False
        elif issubclass(t, bool):
            if isinstance(v, str):
                v = v.lower()
                if v[0] == 'y':
                    ret = True, True
                elif v[0] == 'n':
                    ret = False, True
                elif v[0] == 't':
                    ret = True, True
                elif v[0] == 'f':
                    ret = False, True
                elif v[0] == '1':
                    ret = True, True
                elif v[0] == '0':
                    ret = False, True
                else:
                    ret = v, False
            else:
                ret = bool(v), True
        else: ret = v, False
        if ret[0] in limits: return ret[0], False
        else: return ret
    def end(self): pass
    def doprint(self, *vals): pass
    def can_play_card(self, current_card, card):
        return (card.color == WILD or
        card.number == current_card.number or
        card.color == current_card.color)
    def can_play(self, current_card):
        for card in self.hand:
            if self.can_play_card(current_card, card):
                return True
        else: return False
class RealPlayer(Player):
    def __init__(self, card_count=7):
        super().__init__(card_count)
        self.name = input('What is your name? ')
    def _play(self, current_card):
        self.doprint('Current card:', current_card)
        while True:
            self.doprint(*self.hand)
            hand_colors = sort_cards(self.hand, 'code')
            desired_color = '!'
            while not desired_color or desired_color.lower()[0] not in hand_colors:
                desired_color = input('What color do you want to play? ')
            desired_color = desired_color.lower()[0]
            self.doprint(*hand_colors[desired_color])
            color_numbers = []
            for card in hand_colors[desired_color]:
                color_numbers.append(str(card.number))
            desired_number = ''
            while desired_number not in color_numbers and not desired_number.startswith('esc'):
                desired_number = input('What number do you want to play? ')
                desired_number = desired_number.strip().lower()
                # try: desired_number = int(desired_number)
                # except ValueError: desired_number = ''
            if desired_number.startswith('esc'): continue
            for card in hand_colors[desired_color]:
                if str(card.number) == desired_number: break
            return card
    def _ask(self, q, t):
        return input(q)
    def doprint(self, *vals): print(*vals)
class ComputerPlayer(Player):
    NAMES = [
        'Hal',
        'Cortana',
        'Alexa',
        'Bixby',
        'Siri'
    ]
    used_names = set()
    name_loop_count = 0
    def __init__(self, card_count=7):
        super().__init__(card_count)
        if self.used_names == set(self.NAMES):
            self.used_names.clear()
            ComputerPlayer.name_loop_count += 1
        namelist = list(set(self.NAMES) - self.used_names)
        self.name = random.choice(namelist)
        self.used_names.add(self.name)
        name_id = self.name_loop_count
        if name_id > 0:
            name_id = name_id + 1
        self.name += ' ' + int_to_roman(name_id)
        self.name = self.name.strip()
    @staticmethod
    def tally_special(cardlist):
        value = tally(cardlist)
        multiple = 0.5 * len(cardlist) + 0.5
        return value * multiple
    @staticmethod
    def choose_color(cardlist):
        cards = sort_cards(cardlist)
        cols = list(cards)
        cols.sort(key=(lambda x: ComputerPlayer.tally_special(cards[x])))
        return cols[0]
    def _ask(self, q, t):
        if q == questions['wild']:
            return self.choose_color(self.hand)
        else: return super()._ask(q, t)
    def _play(self, current_card):
        cards = sort_cards(self.hand)
        cols = list(cards)
        playable_cols = set()
        for color in cols:
            for card in cards[color]:
                if self.can_play_card(current_card, card):
                    playable_cols.add(color)
        # colored_cards = [cards[x] for x in playable_cols]
        colored_cards = []
        for color in playable_cols:
            colored_cards += cards[color]
        color = self.choose_color(colored_cards)
        col_cards = cards[color]
        col_cards.sort(key=(lambda x: x.points), reverse=True)
        for card in col_cards:
            if self.can_play_card(current_card, card):
                return card
    def end(self):
        self.used_names.clear()
        ComputerPlayer.name_loop_count = 0

def draw(count):
    hand = nrandom.choice(CARD_LIST, count, False, WEIGHT_LIST)
    return list(hand)

# __all__ = dir()


from netsc import Client
from .other import deserialize_card, serialize_card, GameSubstitute

class GameClient(Client):
    def deadj_args_start(self, *args):
        return (GameSubstitute(self),), {}
    def deadj_args_play(self, *args):
        return (deserialize_card(args[0]), GameSubstitute(self)), {}
    def adj_return_play(self, card):
        if card is None:
            return card
        return serialize_card(card)
    def deadj_args_can_play_card(self, card1, card2):
        return (deserialize_card(card1), deserialize_card(card2)), {}
    def deadj_args_can_play(self, card):
        return (deserialize_card(card),), {}
    def deadj_args_remove_from_hand(self, card):
        return (deserialize_card(card),), {}
    def adj_return___getattribute__(self, value):
        if value is self.wrapped.hand:
            return [serialize_card(x) for x in value]
        else: return value


from ._mods import *
from ._player import *

class Game:
    def __init__(self, players):
        self.players = players
        self.ix = random.randrange(len(self.players))
        self.player = players[self.ix]
        self.card = draw(1)[0]
        self.direction = 1
    def next_player(self):
        self.ix = (self.ix + self.direction) % len(self.players)
        self.player = self.players[self.ix]
    def display_message(self, *vals):
        for player in self.players:
            player.doprint(*vals)
    def game_over(self):
        for player in self.players:
            player.end()
    def begin(self):
        for player in self.players:
            player.start(self)
        while True:
            card = self.player.play(self.card, self)
            if card is None: self.next_player()
            else:
                self.card = card
                if card in self.player.hand:
                    self.player.remove_from_hand(card)
                if not len(self.player.hand): break
                card.played(self)
                self.next_player()
        player_scores = self.players[:]
        player_scores.sort(key=(lambda x: x.score()), reverse=True)
        for (i, player) in enumerate(player_scores[:-1]):
            self.display_message(player.name, 'came in', ordinal(len(self.players)-i),
            'place with', player.score(), 'points.')
        self.display_message(self.player.name, 'wins!')

# __all__ = dir()


from .. import CARD_SET

def serialize_card(card):
    return (card.__class__.__name__, card.short_name, card.color.name)

def deserialize_card(data):
    for card in CARD_SET:
        if data == serialize_card(card):
            return card
    print('ERROR: could not deserialize card, substituting random card')
    return card

class GameSubstitute:
    def __init__(self, client_obj):
        self.client_obj = client_obj
    def __getattr__(self, attr):
        return getattr(self.client_obj, attr)


import sys
_failed_libs = []

try: import numpy.random as nrandom
except ImportError: _failed_libs.append('numpy')

if sys.platform[:3] == 'wi#n':
    import winreg
    reg = winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER)
    try: k = winreg.OpenKey(reg, 'Console\\VirtualTerminalLevel')
    except OSError:
        # k = winreg.CreateKey(reg, 'Console\\VirtualTerminalLevel')
        v = input('ANSI support is not enabled on your system.'
        ' ANSI is used for displaying card colors. Would you like to enable it? ')
        v = v.lower()
        if v[0] == 'y':
            ret = True
        elif v[0] == 'n':
            ret = False
        elif v[0] == 't':
            ret = True
        elif v[0] == 'f':
            ret = False
        elif v[0] == '1':
            ret = True
        elif v[0] == '0':
            ret = False
        else:
            ret = False
        if ret:
            # winreg.OpenKey(reg, 'Console\\VirtualTerminalLevel', winreg.REG_DWORD, winreg.KEY_WRITE)
            winreg.SetValue(reg, 'Console\\VirtualTerminalLevel', winreg.REG_DWORD, '1')
            print('Done')
        else:
            print('You have chosen not to enable it. Cards may not render properly.')
if _failed_libs:
    _libmsg = ', '.join(_failed_libs)
    import sys
    if hasattr(sys, 'uno_exec'):
        _libres = input('The following libraries need to be downloaded and instaled: %s.'
        ' Would you like to do that now? (y,n) ' % _libmsg)
        if _libres in ('Y', 'y'):
            import subprocess, sys
            try:
                subprocess.check_call([sys.executable, '-m', 'pip', 'install', '--user', '--upgrade', 'pip', '--no-warn-script-location'])
            except subprocess.CalledProcessError:
                try:
                    subprocess.check_call(['sudo', 'yum', 'install', 'python3-pip'])
                except subrocess.CalledProcessError:
                    subprocess.check_call(['sudo', 'apt-get', 'install', 'pip3'])
            for _lib in _failed_libs:
                subprocess.call([sys.executable, '-m', 'pip', 'install', '--user', '--upgrade', _lib, '--no-warn-script-location'])
            print('Success!')
            
            import numpy.random as nrandom
        else: import sys; sys.exit()
    else: raise ImportError('The following libraries were unavailable: %s.' % _libmsg)

import math, random, socket, pickle, time, _thread, string, copy
ordinal = (lambda n: "%d%s" % (n,"tsnrhtdd"[(math.floor(n/10)%10!=1)*(n%10<4)*n%10::4]))

# __all__ = dir()


from .server import GameServer
from .client import GameClient

def NetworkPlayer(card_count=7):
    server = GameServer(None)
    addr = server.accept()
    print('%s joined the game!' % (addr,))
    return server


from ._color import *
from ._card import *
from ._colors import *

CARD_SET = set()
COLOR_SET = {RED, GREEN, BLUE, YELLOW, WILD}

for color in ('red', 'green', 'blue', 'yellow'):
    for number in range(1, 10):
        exec("%s_%i = Card('%s %i', '%i', %s, %i, 2, %i)" %
            (color.upper(), number, color.capitalize(), number, number, color.upper(), number, number))
        CARD_SET.add(eval('%s_%i' % (color.upper(), number)))

    exec("%s_0 = Card('%s 0', '0', %s, 0, 1, 0)" % (color.upper(), color.capitalize(), color.upper()))
    CARD_SET.add(eval('%s_0' % color.upper()))

    exec("%s_SKIP = Skip('%s Skip', %s)" % (color.upper(), color.capitalize(), color.upper()))
    CARD_SET.add(eval('%s_SKIP' % color.upper()))

    exec("%s_REVERSE = Reverse('%s Reverse', %s)" % (color.upper(), color.capitalize(), color.upper()))
    CARD_SET.add(eval('%s_REVERSE' % color.upper()))

    exec("%s_DRAW2 = Draw2('%s Draw 2', %s)" % (color.upper(), color.capitalize(), color.upper()))
    CARD_SET.add(eval('%s_DRAW2' % color.upper()))
WILD_CARD = Wild()
WILD_DRAW4 = WildDraw4()
CARD_SET.add(WILD_CARD)
CARD_SET.add(WILD_DRAW4)

def calculate_chance():
    weight_total = 0
    cards = []
    chances = []
    for card in CARD_SET:
        cards.append(card)
        weight_total += card.weight
    for card in CARD_SET:
        chances.append(card.weight / weight_total)
    return cards, chances
CARD_LIST, WEIGHT_LIST = calculate_chance()

# __all__ = dir()


from ._mods import *
from ._color import *
from ._colors import *

class Card:
    def __init__(self, long_name, short_name, color, number, weight, points):
        self.long_name = long_name
        self.short_name = short_name
        self.color = color
        self.number = number
        self.weight = weight
        self.points = points
    def played(self, game): pass
    def __str__(self):
        return '\u001b[%sm%s\u001b[37m' % (self.color.ansi, self.short_name)
    def __repr__(self):
        return '[%s: %s]' % (self.__class__.__name__, self.long_name)

class Skip(Card):
    def __init__(self, long_name, color):
        super().__init__(long_name, 'S', color, 's', 2, 20)
    def played(self, game):
        game.ix += 1
        game.display_message("%s has been skipped"
        % game.players[game.ix%len(game.players)].name)
        # % game.players[(game.ix+1)%len(game.players)].name)
class Reverse(Card):
    def __init__(self, long_name, color):
        super().__init__(long_name, 'R', color, 'r', 2, 20)
    def played(self, game):
        if len(game.players) < 3:
            game.ix += 1
            game.display_message("%s has been skipped"
            % game.players[game.ix%len(game.players)].name)
        else:
            game.direction = -game.direction
            game.display_message("%s has reversed the direction" % game.player.name)
class Draw2(Card):
    def __init__(self, long_name, color):
        super().__init__(long_name, 'D2', color, 'd2', 2, 20)
    def played(self, game):
        game.ix += 1
        print("%s forced %s to draw two cards"
        % (game.player.name,
        game.players[game.ix%len(game.players)].name))
        game.players[game.ix%len(game.players)].draw(2)

class Wild(Card):
    def __init__(self, long_name='Wild', short_name='W', number='w'):
        super().__init__(long_name, short_name, WILD, number, 4, 50)
    def played(self, game):
        game.card = copy.copy(self)
        game.card.color = game.player.ask(
            questions['wild'], Color, limits=(WILD,))
        print('%s changed the color to %s.' % (game.player.name, game.card.color.name))
class WildDraw4(Wild):
    def __init__(self):
        super().__init__('Wild Draw 4', 'D4', 'd4')
    def played(self, game):
        cur_color = game.card.color
        super().played(game)
        game.ix += 1
        attacked_player = game.players[game.ix%len(game.players)]
        print("%s forced %s to draw four cards."
        % (game.player.name, attacked_player.name))
        called = attacked_player.ask(questions['call wild'], bool)
        if called:
            if cur_color in sort_cards(game.player.hand):
                print("%s called and forced %s to draw four cards instead."
                % (attacked_player.name, game.player.name))
                game.player.draw(4)
                game.ix -= 1
            else:
                print("%s called and was wrong, so they had to draw six cards."
                % attacked_player.name)
                attacked_player.draw(6)
        else: attacked_player.draw(4)

def get_card_name(card):
    value = card.long_name.upper().replace(' ', '_')
    prevchar = ''
    newval = ''
    reached__ = False
    for char in value:
        if reached__ and prevchar == '_' and char in string.digits:
            newval = newval[:-1]
        if prevchar == '_' and not reached__:
            reached__ = True
        newval += char
        prevchar = char
    return newval
def tally(seq):
    res = 0
    for card in seq:
        res += card.points
    return res
def sort_cards(seq, n='object'):
    if n == 'object':
        do = compile('card.color', '<card_sort>', 'eval')
    elif n == 'code':
        do = compile('card.color.code', '<card_sort>', 'eval')
    res = {}
    for card in seq:
        if eval(do) not in res:
            res[eval(do)] = []
        res[eval(do)].append(card)
    return res

from ._game import *
_r_core = False
if not _r_core:
    from ._core import *
    _r_core = True

# __all__ = dir()